// by: Scott R. Gravenhorst
// 2011-02-01
//
// This is a dual mode filter module with selectable single pole LPF and single pole HPF.
// Both filters execute in 3 clocks.
//
// Highpass math:
// Start with y[n] = (1+bw)/2 * x[n] + (-1+bw)/2 * x[n-1] + bw * y[n-1]    -- the standard form.
//   To reduce the number of multiplies from 3 to 2, algebra produces:
//            y[n] = (1/2+bw/2) * (x[n] - x[n-1]) + bw * y[n-1]
//   note that the form (1+bw)/2 converted to (1/2+bw/2) allows the use of fixed point binary math
//   because there are no numbers >= 1.00000.  i.e., 1+bw can be larger than one, dividing by 2 first
//   means the two values summed 1/2 and bw/2 are both <= 0.5.
//
//  for this synth (tonal noise synth) HP is disabled since this filter is used to generate portamento (LP)

module single_pole_IIR(clk, ena, bw, in, sel, type, out);
  parameter SIZE = 35;
  parameter HIBIT = SIZE - 1;
  parameter MULT_SIZE = SIZE * 2;
  
  input                        clk;                // System clock
  input                        ena; 
  input signed  [HIBIT:0]      bw;                 // bandwidth.
  input signed  [HIBIT:0]      in;                 // Input data
  input         [4:0]          sel;                // select one of 32 filters
  input                        type;               // 0 = lowpass / 1 = highpass
  output signed [HIBIT:0]      out;                // Output data

  wire                         clk;
  wire                         ena;
  wire signed   [HIBIT:0]      bw;
  wire signed   [HIBIT:0]      in;                 // Input data
  wire          [4:0]          sel;                // select one of 32 filters
  wire                         type;               // filter type, lowpass or highpass
  wire signed   [HIBIT:0]      out;                // Output data
  
  wire signed   [HIBIT:0]      MAX_VAL;
  assign MAX_VAL = {1'b0,{SIZE-1{1'b1}}} ;         // maximum fixed point value
  
  wire signed   [HIBIT:0]      LP_a0;
  wire signed   [HIBIT:0]      LP_b1;

//  assign LP_a0 = {1'b0,bw,{SIZE-22{1'b0}}};        // a0 for lowpass computed from bw
  assign LP_a0 = bw;                               // a0 for lowpass computed from bw
  assign LP_b1 = MAX_VAL - bw;                     // b1 for lowpass computed from bw

//  reg  signed   [HIBIT:0]      PREV_in = 0;

  reg signed    [HIBIT:0]      Z1 = 0;             // Z1 memory/delay cell
  reg signed    [HIBIT:0]      Z1_RAM [31:0];      // 32 Z1 delay elements, dist RAM
//  reg signed    [HIBIT:0]      PREV_in_RAM [31:0];  // used only for HP

  reg signed    [HIBIT:0]      iirA = 0;           // multiplier A register
  reg signed    [HIBIT:0]      iirB = 0;           // multiplier B register
  wire signed   [MULT_SIZE-1:0] iirP;              // output of multiplier
  wire signed   [HIBIT:0]      iirPROD;            //
  reg           [HIBIT:0]      iirPROD_REG = 0;    // output holding register

  assign iirP = iirA * iirB ;                      // instantiate a multiplier
  assign iirPROD = ( iirP >>> (SIZE-1) ) ;         // restore value to SIZE bits
  
  assign out = Z1_RAM[sel];                        // both filters output

///////////////////////////////////////////////////////////////////////////////////////
// filters state machine:

  reg [1:0] state = 0;
  reg       run = 0;

  always @ ( posedge clk )
    begin
    if ( ena )
      begin
      run         <= 1;
      state       <= {1'b0,type};                      // set state based on whether highpass(1) or lowpass(0)
      
      Z1          <= Z1_RAM[sel];                      // Z1 is used by both highpass and lowpass
      
//      if ( type == 0 )
//        begin
        iirA        <= in;                             // lowpass data input
        iirB        <= LP_a0;                          // lowpass a0 value
//        end
//      else
//        begin
//        iirA        <= ( in - PREV_in_RAM[sel] ) ;     // highpass data input
//        iirB        <= 28'sh4000000 + ( bw >>> 1 );    // 0.5 + bw/2
//        end
      end
      
    else
    
      begin
      if ( run )
        begin
        case ( state )

        // LPF:
          2'b00:                                       // this state is used only for lowpass filter
            begin  state <= 2'b10;                     // state 2'b10 is the finish state for both filters
            iirPROD_REG <= iirPROD;                    // save first product for summing later
            iirA        <= Z1;                         // set up for Z1 * b1
            iirB        <= LP_b1;                      //
            end
          
        // HPF:
//          2'b01:                                       // this state is used only for highpass filter
//            begin  state <= 2'b10;                     // state 2'b10 is the finish state for both filters
//            iirPROD_REG <= iirPROD;                    // save first produce for summing later
//            iirA        <= Z1;                         // set up for Z1 * bw
//            iirB        <= bw;                         // 
//            end
            
          2'b10:
            begin
            run <= 0;
//            PREV_in_RAM[sel] <= in;                    // matters only for HPF
            Z1_RAM[sel] <= (iirPROD + iirPROD_REG);    // sum and store output for use in next cycle
            end

        endcase
        end
      end      
    end

endmodule
