`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Engineer: Scott R. Gravenhorst
// email: music.maker@gte.net
// Create Date:    01/26/2010
// Design Name:    SVF
// Module Name:    SVF
// Project Name:   State Variable Filter Array
// Description:    SVF with shared multiplier.
// 
// Using 35 bit signed arithmetic to expand Q
// Gain due to Q accomodated by allowing only signed 24 bit input while supplying
// a signed 35 bit output.
//
// Execution time = 9 clocks
//
// 32 filters, RAM-based
//
// uses 4 dedicated multipliers
//
// - Improved timing
// - Added reset to squelch input to RAMs.  This should allow silencing an unstable
//   filter.
// - 3 outputs, lowpass, highpass and bandpass
// - Adding Overflow detection by sacrificing one output bit to be used as a second
//   sign bit.
// - State machine updated with an output register and output RAM which allows the 
//   last cycle's computed output to be collected one clock after ena.
// - add register to store sel value so that it cannot be influenced by changes 
//   outside this module.
// - Improved implementation of overflow detection by doing a 1 position right shift 
//   of the input data and left shifting 1 position for output.  Detection is then
//   a matter of XORing the two most significant bits anywhere you need to worry about
//   it.
// 
//////////////////////////////////////////////////////////////////////////////////

module SVF( clk, ena, sel, f, q, In, SVFtyp, Out, oflo, reset );

  parameter DATAWIDTH = 35;             // 35 bit signed arithmetic
  parameter HIBIT = DATAWIDTH - 1;
  parameter MULT_OUT_HIBIT = (2*DATAWIDTH)-1;

  input clk;                            // system clock
  input ena;                            // Tell the filter to go
  input         [4:0]     sel;          // select one of 32 filters
  input signed  [HIBIT:0] f;            // f (not Hz, but usable to control frequency)
  input signed  [HIBIT:0] q;            // q (1/Q)
  input signed  [HIBIT:0] In;           // 24 bit signed input
  input         [1:0]     SVFtyp;       // select one of four outputs LP, HP, BP, BR
  output signed [HIBIT:0] Out;
  output                  oflo;
  input                   reset;
  
  wire clk;
  wire ena;
  wire         [4:0]     sel;
  wire signed  [HIBIT:0] f;
  wire signed  [HIBIT:0] q;
  wire signed  [HIBIT:0] In;
  wire         [1:0]     SVFtyp;
  reg  signed  [HIBIT:0] Out;
  reg                    oflo = 0;
  wire                   reset;
  
  reg         [4:0]  sel_REG = 0;
  reg         [1:0]  SVFtyp_REG = 0;
  reg signed  [HIBIT:0] InReg;

///////////////////////////////////////////////////////////////////////////////////////
// SVF RAM for feedback storage  

  reg signed  [69:0] Z1_I;             // RAM input register  -  extra bit for overflow detection

  reg signed  [69:0] Z2_I;             // RAM input register  -  extra bit for overflow detection

  reg [69:0] Z1_RAM [31:0];
  reg [69:0] Z2_RAM [31:0];

  reg [HIBIT:0] Out_RAM [31:0];       // stores each filter's output
  
///////////////////////////////////////////////////////////////////////////////////////  

  reg  signed [HIBIT:0] OutMux;                       // lowpass output

  reg  signed [HIBIT:0] mA = 0;
  reg  signed [HIBIT:0] mB = 0;
  wire signed [MULT_OUT_HIBIT:0] mP;                // multiplier output P
  reg  signed [MULT_OUT_HIBIT:0] mP_REG;
  assign mP = mA * mB;                              // shared multiplier

// SVF state machine, shares a multiplier
  reg signed [69:0] tmp;

  reg       run   = 0;
  reg [2:0] state = 0;

  always @ ( posedge clk )
    begin

    if ( ena ) 
      begin
      run <= 1'b1;
      state <= 3'd0;

      if ( reset )  oflo <= 0; 
      else if ( Out[HIBIT] != Out[HIBIT-1] ) oflo <= 1; // overflow detection
      
      Out   <= Out_RAM[sel] << 1;  // Supply last cycle's output, compensate for >>> 1 of input for overflow check
      InReg <= In >>> 1;           // Get the input value, >>> 1 to implement overflow detection

      SVFtyp_REG <= SVFtyp;
      sel_REG <= sel;
      
      mA    <= f;
      mB    <= (Z1_RAM[sel] >>> 34);
      end

    else

      begin
      if ( run )
        begin
        state <= state + 1;

        case ( state )

          3'd0:
            begin
            mP_REG <= mP;
            end              

          3'd1:
            begin
            tmp <= ( ( InReg << 35 ) - Z2_RAM[sel_REG] );       // part of first summing node

            Z2_I <= (reset) ? 0 : (mP_REG + Z2_RAM[sel_REG]);   // mP = f * (Z1_O >>> 34)

            mA <= q;
            mB <= (Z1_RAM[sel_REG] >>> 34);                     // prepare for q * z1
            end

          3'd2:
            begin
            mP_REG <= mP;
            end              

          3'd3:
            begin
            mA <= f;
            mB <= ( tmp - mP_REG ) >>> 34;                      // mB is also the HP output
            end
          
          3'd4:
            begin
            mP_REG <= mP;                
            end

          3'd5:
            begin
            Z1_I <= (reset) ? 0 : (mP_REG + Z1_RAM[sel_REG]);  // mP = f * (((In35 << 35) - mP - Z2_O) >>> 34) 
            end

          3'd6:
            begin
            Z1_RAM[sel_REG] <= Z1_I;
            Z2_RAM[sel_REG] <= Z2_I;
            
            case ( SVFtyp_REG )  // Output mux    00 = LP   01 = HP   10 = BP    11 = BR
              2'h0: OutMux <= ( Z2_RAM[sel_REG] >>> 35 );      // lowpass output
              2'h1: OutMux <= mB;                          // highpass output
              2'h2: OutMux <= ( Z1_I >>> 35 );             // bandpass output
              2'h3: OutMux <= mB + ( Z2_RAM[sel_REG] >>> 35 ); // bandreject output
            endcase
            end
            
          3'd7:
            begin
            Out_RAM[sel_REG] <= OutMux;

            run <= 1'b0;                                   // stop state machine
            end
            
        endcase

        end 
      end
    end
  endmodule
