`timescale 1ns / 1ps
// Engineer: Scott R. Gravenhorst
// email: music.maker@gte.net
// Date Started: 2011-02-12
// Design Name: Tonal Noise Monosynth with Portamento and inverse Q scaling
// Description: MIDI triggered digital monosynth.
//
//
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Tonal Noise Monosynth
//
// ver_a - First cut and proof of concept.
//         Works - including Q enhancement compensation.
//       - Add amplitude compensation to make amplitude more consistent over the range of Fc.
//         Currently, the amplitude is far less at the low end than at the high end.  Seems to help
//      
// ver_b - Fix SVF code, the module should accept a 35 bit signed input and supply a 35 bit signed
//         output.  Internally it will >>> 1 position to implement overflow detection.  Thus the SVF
//         will operate as signed 34 bit. -- works very well.
//
// ver_c - Add second BP filter with fixed Q of 1, but with Fc the same as the first.  Hopefully,
//         this will reduce the amount of noise away from the desired pitch.
//         Stretch state machine to 48 states moving output logic to the end. 
//
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

module Tonal_Noise ( CLK_50MHZ, LED,
                   lcd_rs, lcd_rw, lcd_e, lcd_d, 
                   BTN_E, BTN_W,
                   spi_sck, spi_dac_cs, spi_sdi, spi_rom_cs, 
                   spi_amp_cs, spi_adc_conv, spi_dac_cs, spi_amp_shdn, spi_dac_clr,
                   strataflash_oe, strataflash_ce, strataflash_we,
                   platformflash_oe,
                   Raw_MIDI_In,
                   TTY_In
                   );

// This parameter supplies the version number to an MCU port which is displayed in the LCD.
/////////////////////////////////////////////////////////////////////////////////////////
                              //                                                       //
                              //    #       #  #####  ####    ###   ###   ###   #   #  //
                              //     #     #   #      #   #  #       #   #   #  ##  #  //
  parameter version = "c";    //      #   #    ###    ####    ###    #   #   #  # # #  //
                              //       # #     #      #  #       #   #   #   #  #  ##  //
                              //        #      #####  #   #   ###   ###   ###   #   #  //
                              //                                                       //
/////////////////////////////////////////////////////////////////////////////////////////

  parameter ARITH_WIDTH = 35;
  parameter HIBIT = ARITH_WIDTH - 1;

  input CLK_50MHZ;
  output [7:0] LED;
  
  inout [7:4] lcd_d;
  output lcd_rs;
  output lcd_rw;
  output lcd_e;
  
  input BTN_E;     // MCU reset
  input BTN_W;     // master tune multiplier, when pushed, causes increment of 16 instead of 1.
  
  output spi_sck;
  output spi_sdi;
  output spi_rom_cs;
  output spi_amp_cs;
  output spi_adc_conv;
  output spi_dac_cs;
  output spi_amp_shdn;
  output spi_dac_clr;
  output strataflash_oe;
  output strataflash_ce;
  output strataflash_we;
  output platformflash_oe;

  input Raw_MIDI_In;
  input TTY_In;

  wire CLK_50MHZ;
  wire clk;

  wire [7:4] lcd_d;
  wire lcd_rs;
  wire lcd_rw;
  wire lcd_rw_control;
  wire lcd_e;
  wire lcd_drive;
  
  reg [7:0] LCD;
  
  assign lcd_d[7:4] = ( lcd_drive == 1'b1 & lcd_rw_control == 1'b0 ) ? LCD[7:4] : 4'bzzzz;
  assign lcd_drive = LCD[3];
  assign lcd_rs = LCD[2];
  assign lcd_rw_control = LCD[1];
  assign lcd_e = LCD[0];
  assign lcd_rw = lcd_rw_control & lcd_drive;

  reg [11:0] DACreg = 12'b000000000000;  // used by SPI DAC

  wire [7:0] LED;

  wire DACena;

  wire BTN_E, BTN_W;

  wire [3:0] rstd;                // POR delay
  wire reset;                     // POR/User reset
  
  wire interrupt;
  wire interrupt_ack;
  wire interrupt0;
  wire interrupt1;
  wire [9:0] address;             // wires to connect address lines from uC to ROM
  wire [17:0] instruction;        // uC data lines, need connection between uC and ROM
  wire [7:0] out_port;            //
  wire [7:0] in_port;             // 
  wire [7:0] port_id;
  wire proc_reset;                // an output from the mcu ROM, not sure what it's supposed to connect to.
  reg [7:0] in_port_reg;          // hold data for mcu
  assign in_port = in_port_reg;

// MIDI & TTY UART receivers
  wire [7:0] rx0data;
  wire rx0ready_status;
  wire reset_rx0ready_status;
  
  wire [7:0] rx1data;
  wire reset_rx1ready_status;

  wire resetsignal;
// MCU
  wire read_strobe;
  wire write_strobe;

///////////////////////////////////////////////////////////////////////////////////////  
// DCM

 DCM_40 DCM ( .CLKIN_IN( CLK_50MHZ ), .CLKFX_OUT( clk ), .CLKIN_IBUFG_OUT(), .CLK0_OUT() );

///////////////////////////////////////////////////////////////////////////////////////  
// Synth signals

//  wire [11:0] DACbus;

  reg [3:0] MIDI_CHANNEL = 0;

  reg [7:0] SYSEX_ADDR_MSB = 8'h00;

  reg         [7:0] NOTENUM;            // MIDI note number, used as a modulation source
  wire        [7:0] NOTENUMrev;         // 127 - NOTENUM for amplitude compensation
  assign NOTENUMrev = 8'h7F - NOTENUM;

// 7 bit:  (these are supplied by sysex messages, thus are restricted to 7 bits.
// 14 bit values are assembled in hardware.
//  reg [13:0] PW = 14'b10000000000000;  // init pitch wheel at center in case it never sends a message
//  reg [6:0] CHANpres;
//  reg [6:0] MOD_WHL;
//  reg [6:0] VEL;
  reg [6:0] TRANSPOSE = 7'h00;         // global transposition in half steps
  reg       MCU_GATE = 1'b0;
  reg       SUSTAIN = 1'b0;
 
  reg [7:0] NOTEOCT;       // DIST. RAM - port registers to receive note-octave information from MCU

///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
// POR delay FF chain - taken from Eric Brombaugh's code for the SPI DAC.
  FDCE rst_bit0 (.Q(rstd[0]), .C(clk), .CE(1'b1), .CLR(1'b0), .D(1'b1));
  FDCE rst_bit1 (.Q(rstd[1]), .C(clk), .CE(1'b1), .CLR(1'b0), .D(rstd[0]));
  FDCE rst_bit2 (.Q(rstd[2]), .C(clk), .CE(1'b1), .CLR(1'b0), .D(rstd[1]));
  FDCE rst_bit3 (.Q(rstd[3]), .C(clk), .CE(1'b1), .CLR(1'b0), .D(rstd[2]));
  assign reset = ~rstd[3] | BTN_E;    // use east button as reset.

// Tie off the flash enables to allow SPI to work
  assign strataflash_oe = 1'b1;
  assign strataflash_ce = 1'b1;
  assign strataflash_we = 1'b1;
  assign platformflash_oe = 1'b0;
  
// Tie off other SPI enables to isolate DAC
  assign spi_rom_cs = 1'b1;
  assign spi_amp_cs = 1'b1;
  assign spi_adc_conv = 1'b0;
  assign spi_amp_shdn = 1'b1;
  assign spi_dac_clr = 1'b1;
  
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
// instantiate the uC (kcpsm3) with it's ROM

  kcpsm3 MCU ( .address(address), .instruction(instruction), .port_id(port_id), 
    .write_strobe(write_strobe), .out_port(out_port), .read_strobe(read_strobe), .in_port(in_port), 
    .interrupt(interrupt), .interrupt_ack(interrupt_ack), 
    .reset(proc_reset|reset), .clk(clk) );  

  midictrl PSMROM ( .address(address), .instruction(instruction), .proc_reset(proc_reset), .clk(clk) );

///////////////////////////////////////////////////////////////////////////////////////
// MIDI UART
  MIDIuartrx RX0 ( .dout(rx0data), .clk(clk), .reset(resetsignal), .rxd(Raw_MIDI_In), 
    .frame(), .overrun(), .ready(), .busy(), .CS(), 
    .interrupt(interrupt0), .interrupt_ack(interrupt_ack), 
    .rxready_status(rx0ready_status), 
    .reset_rxready_status(reset_rx0ready_status)
    );
  /////// VERY IMPORTANT HARDWARE /////////////////////////////////////////////
  // decode read port 01, send pulse to reset rxready flop
  // This allows the mcu to clear the rxready bit automatically just by reading rxdata.
  assign reset_rx0ready_status = (read_strobe == 1'b1) & (port_id[3:0] == 4'h1);

///////////////////////////////////////////////////////////////////////////////////////
// TTY UART, 115.2 or 19.2 kilobuad (baudrate configured in module)
  TTYuartrx RX1 ( .dout(rx1data), .clk(clk), .reset(resetsignal), .rxd(TTY_In), 
    .frame(), .overrun(), .ready(), .busy(), .CS(), 
    .interrupt(interrupt1), .interrupt_ack(interrupt_ack), 
    .rxready_status(rx1ready_status), 
    .reset_rxready_status(reset_rx1ready_status)
    );
  /////// VERY IMPORTANT HARDWARE /////////////////////////////////////////////
  // decode read port 09, send pulse to reset rxready flop
  // This allows the mcu to clear the rxready bit automatically just by reading rxdata.
  assign reset_rx1ready_status = (read_strobe == 1'b1) & (port_id[3:0] == 4'h9);

// common
  assign interrupt = (interrupt0 | interrupt1);  // ISR gets to figure out which UART did it.
  assign resetsignal = write_strobe && ( port_id == 8'hFF );  // When port_id == FF with write strobe, reset the UARTs

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// DAC, module by Eric Brombaugh, modified by Scott Gravenhorst

  parameter CYCLES = 12'd50;             // SR = 800 KHz at 40 MHz system clock
  
  spi_dac_out DAC (
    .clk( clk ), 
    .reset( reset ), 
    .spi_sck( spi_sck ), 
    .spi_sdo( spi_sdi ), 
    .spi_dac_cs( spi_dac_cs ), 
    .ena_out( DACena ), 
    .data_in( DACreg ), 
    .cycles( CYCLES )               // number of clk cycles per DACena.  50 minimum.
    );
    
  reg [3:0] DACwin = 0;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// Rotary Encoder
               
//  RotaryEncoder_v2 ENC (
//    .clk(clk),
//    .ROTa(ROTa),
//    .ROTb(ROTb),
//    .ROTpress(ROTpress),
//    .BTN_W(BTN_W),
//    .value_out(ROTout),
//    .ROTpress_out(ROTpress_out)
//    );

///////////////////////////////////////////////////////////////////////////////
// Synth component interconnection objects:

  wire GATE;

  wire signed   [HIBIT:0]      MAX_VAL;
  assign MAX_VAL = {1'b0,{HIBIT{1'b1}}} ;         // maximum positive fixed point value

  reg  signed   [HIBIT:0]      AmpComp;           // amplitude compensation

///////////////////////////////////////////////////////////////////////////////
// Sustain pedal logic

  reg SUS = 1'b0;
  assign GATE = MCU_GATE | SUS;             // single bit

  always @ ( posedge clk )
    begin
    case ( SUSTAIN )                        // SUSTAIN is a single bit representing the sustain pedal state
      1'b0: SUS <= 1'b0;                    // if pedal is up, always clear SUS
      1'b1: if ( MCU_GATE ) SUS <= 1'b1;    // if pedal is down, only set SUS on MCU_GATE high
    endcase
    end

///////////////////////////////////////////////////////////////////////////////
// 64 bit LFSR noise generator

  reg [63:0] sr = 64'h461B87AA9928112E;          // 64 millionth iteration after using 64'h0000000000000001 as starting value;
  
  wire signed [HIBIT:0] NZout = sr[HIBIT:0];
  reg  signed [HIBIT:0] NZlev = 0;               // register to control level of noise presented to BP filter
  
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// ADSR, fired twice, once for amplitude and once for Fc of lowpass SVF

  reg [13:0] NCFadsrA = 0;
  reg [13:0] NCFadsrD = 0;
  reg [13:0] NCFadsrS = 0;
  reg [13:0] NCFadsrR = 0;
  reg        expo_R_NCF_ADSR = 0;

  reg [13:0] NCAadsrA = 0;
  reg [13:0] NCAadsrD = 0;
  reg [13:0] NCAadsrS = 0;
  reg [13:0] NCAadsrR = 0;
  reg        expo_R_NCA_ADSR = 0;

  reg         [13:0] ADSR_A = 0;
  reg         [13:0] ADSR_D = 0;
  reg         [13:0] ADSR_S = 0;
  reg         [13:0] ADSR_R = 0;
  reg         [5:0]  ADSRsel = 0;
  reg                ADSRena = 0;
  reg                expo_R = 0;
  wire signed [17:0] ADSRout;
  
  ADSR EG0 (
    .out( ADSRout ), 
    .clk( clk ), 
    .ena( ADSRena ), 
    .sel( ADSRsel ),                 // 1 of 64 
    .GATE( GATE ), 
    .A( ADSR_A ),                    // 14 bits
    .D( ADSR_D ),                    // 14 bits
    .S( {ADSR_S,3'b000} ),           // 17 bits
    .R( ADSR_R ),		                 // 14 bits
    .expo_R( expo_R )
    );

  reg [HIBIT:0] ADSRncf = 0;         // holding register for NCF ADSR
  reg [HIBIT:0] ADSRnca = 0;         // holding register for NCA ADSR

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// NCF - A digital state variable filter with resonance.

  parameter SVF_LP = 0;
  parameter SVF_HP = 1;
  parameter SVF_BP = 2;
  parameter SVF_BR = 3;

  wire                  SVF_oflo;      // overflow state of SVF
  reg         [4:0]     SVFsel = 0;    // 1 of 32
  reg                   SVFena = 0;    // tell filter to go.
  reg         [1:0]     SVFtyp = 0;
  reg  signed [HIBIT:0] SVFin = 0;     // this will be arithmetically scaled
  wire signed [HIBIT:0] SVFout;        // the registered output of the SVF module
  reg  signed [HIBIT:0] SVF_Fc;        // Fc, 35 bit fixed point
  reg  signed [HIBIT:0] SVF_q;         // q, 35 bit fixed point.  q is 1/Q
  
  reg  signed [HIBIT:0] q_Min = 0;     // the value that forms the base q value which will be added to the scaled pitch value to give q for the bandpass filter
  reg  signed [HIBIT:0] q_Scale = 0;   //
  reg  signed [HIBIT:0] Lowpass_q = 0; //

// NOTE: This SVF sacrifices one bit to implement overflow detection.  
// It must be given no more than 1/2 scale amplitude which can be scaled after the output
  SVF SVF (
    .clk( clk ), 
    .ena( SVFena ), 
    .sel( SVFsel ),                    // select SVF0 for the bandpass and SVF1 for the lowpass
    .f( SVF_Fc ),                      // 35 bit fixed point 
    .q( SVF_q ), 
    .In( SVFin ), 
    .SVFtyp( SVFtyp ),                 // selects LP, HP, BP and BR 
    .Out( SVFout ), 
    .oflo( SVF_oflo ), 
    .reset( reset )
    );

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////  
// single pole lowpass IIR
  
  reg                   IIRena = 0;
  reg         [HIBIT:0] Pitch = 0;
  reg         [HIBIT:0] PmentoRate = 0;
  wire signed [HIBIT:0] PmentoOut;
  
  single_pole_IIR IIR (
    .clk( clk ),       //
    .ena( IIRena ),       //
    .bw( PmentoRate ), // bandwidth value, higher is faster
    .in( Pitch ),      // filter input
    .sel( 5'b00000 ),  // force 0, only one filter needed
    .type( 1'b0 ),     // force lowpass
    .out( PmentoOut )  // output as portamento affected pitch
    );

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Tuning ROM
  
  wire [17:0] tunROM;                              // tuning ROM output
  TuningROM TUN ( .O(tunROM), .A(NOTEOCT[3:0]) );

  reg signed [20:0] Tuning;                        // a fixed point value multiplied by the output of the tuning ROM
  
//////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Shared 35x35 multiplier

  parameter MULT_OUT_HIBIT = (2*ARITH_WIDTH)-1;
  
  reg   signed [HIBIT:0] A0;           // multiplier input A0
  reg   signed [HIBIT:0] B0;           // multiplier input B0
  reg   signed [HIBIT:0] PROD0_REG = 0;// temp register to hold product
  wire  signed [HIBIT:0] PROD0;        // 35 bit output
  wire  signed [MULT_OUT_HIBIT:0] P0;  // multiplier output P0, full precision

  assign P0 = A0 * B0;                 //
  assign PROD0 = (P0 >>> HIBIT);       // Get most significant 35 bits with sign

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////  
// State Machine

reg [34:0] DIAG;

  reg        [5:0]  state = 0;
  reg               run = 0;
  reg signed [11:0] SynthOut = 0;        // output register from synth

  always @ ( posedge clk )
    begin
    if ( DACena )
      begin
      state <= 0;
      run   <= 1;
      
      DACreg <= SynthOut + 12'h800;
      end

    else // if NOT DACena

      begin
      if ( run )
        begin
        
        state <= state + 1;
        
        case ( state )
        6'h00:
          begin
          sr <= {sr[62:0],(( sr[63] ^ sr[62] ) ^ ( sr[60] ^ sr[59] ))} ;  // LFSR XOR logic, use NZout for output

          Pitch   <= ( {9'b000000000,tunROM,8'b00000000} << NOTEOCT[7:4] ); // upper 4 bits of NOTEOCT are the octave.

          A0 <= {1'b0,Tuning,13'h0000};
          B0 <= PmentoOut;                  // output of IIR LP 1 pole filter

          IIRena <= 1;
          end

        6'h01:
          begin
          A0 <= NZout;
          B0 <= NZlev;

          SVF_Fc <= PROD0;                  // set to desired pitch

          IIRena <= 0;
          end
          
        6'h02:
          begin                             // SVF_Fc is already set
          SVFin  <= PROD0;                  // 1st BPF input is level modulated noise
          SVF_q  <= 35'sh3FFFFFFFF;         // set Q to 1.0000000000582076609168555390158
          SVFsel <= 2;                      // SVF2
          SVFtyp <= SVF_BP;                 // Bandpass
          SVFena <= 1;                      // filter go
          end
          
        6'h03:                              // SVF2 state 0 complete
          begin
          ADSR_A  <= NCFadsrA;
          ADSR_D  <= NCFadsrD;
          ADSR_S  <= NCFadsrS;
          ADSR_R  <= NCFadsrR;
          expo_R  <= expo_R_NCF_ADSR;
          ADSRsel <= 0;
          ADSRena <= 1;

          SVFena  <= 0;
          SVFin   <= SVFout;                // input to 2nd bandpass filter is output of 1st bandpass filter
          end

        6'h04:                              // SVF2 state 1 complete
          begin
          ADSRncf <= {ADSRout,17'h00000};   // store ADSR for NCF output value
          ADSRena <= 0;
          end          

        6'h05:                              // SVF2 state 3 complete
          begin
          end

        6'h06:                              // SVF2 state 4 complete
          begin
          end

        6'h07:                              // SVF2 state 5 complete
          begin
          end

        6'h08:                              // SVF2 state 6 complete
          begin
          end

        6'h09:                              // SVF2 state 7 complete
          begin
          end

        6'h0A:                              // SVF2 state 8 complete
          begin
          A0 <= ( SVF_Fc << 1 );            // Q enhancement compensation  - q_Scale * SFV_Fc * 2
          B0 <= q_Scale;
          end

        6'h0B:                              // IIR state 2 complete, ADSR state 2 complete
          begin
          PROD0_REG <= PROD0;               // PROD0 is SVF_Fc * q_Scale * 2 (compensation product)
          
          ADSR_A <= NCAadsrA;
          ADSR_D <= NCAadsrD;
          ADSR_S <= NCAadsrS;
          ADSR_R <= NCAadsrR;
          expo_R <= expo_R_NCA_ADSR;
          ADSRsel <= 1;
          ADSRena <= 1;
          end

        6'h0C:
          begin
          SVFsel <= 0;
          SVF_q  <= PROD0_REG + q_Min;     // Q enhancement compensation added to minimum q
          SVFtyp <= SVF_BP;
          SVFena <= 1;
          
          ADSRena <= 0;
          ADSRnca <= {ADSRout,17'h00000};   // store ADSR for NCA output value
          end

        6'h0D:                              // SVF0 state 0 complete
          begin
          SVFena <= 0;
          
          SVFin <= SVFout;                  // set up for lowpass with bandpass as input
          end

        6'h0E:                              // SVF0 state 1 complete
          begin
          end

        6'h0F:                              // SVF0 state 2 complete
          begin
          end

        6'h10:                              // SVF0 state 3 complete
          begin
          end

        6'h11:                              // SVF0 state 4 complete
          begin
          end

        6'h12:                              // SVF0 state 5 complete
          begin
          end

        6'h13:                              // SVF0 state 6 complete
          begin
          end

        6'h14:                              // SVF0 state 7 complete
          begin
          end

        6'h15:                              // SVF0 state 8 complete
          begin
          A0 <= AmpComp;
          B0 <= {NOTENUMrev,27'h7FFFFFF};
          
//          B0 <= (MAX_VAL - ((SVF_Fc - 35'sh0_01F0_0000) << 1));

          SVFsel <= 1;                      // SVF1
          SVFtyp <= SVF_LP;                 // Lowpass
          SVF_Fc <= ADSRncf;                // get Fc from NCF ADSR
          SVF_q  <= Lowpass_q;              // set q
          SVFena <= 1;                      // filter go
          end

        6'h16:                              // SVF1 state 0 complete
          begin
          SVFena <= 0;                      //

          A0 <= SVFout;                     // lowpass SVF output
          B0 <= PROD0;                      // (MAX_VAL - (SVF_Fc << 1)) * AmpComp
          end
          
        6'h17:                              // SVF1 state 1 complete
          begin                             // PROD0 has the compensated first SVF output
          end

        6'h18:                              // SVF1 state 2 complete
          begin
          end

        6'h19:                              // SVF1 state 3 complete
          begin
          end

        6'h1A:                              // SVF1 state 4 complete
          begin
          end

        6'h1B:                              // SVF1 state 5 complete
          begin
          end

        6'h1C:                              // SVF1 state 6 complete
          begin
          end

        6'h1D:                              // SVF1 state 7 complete
          begin
          end

        6'h1E:                              // SVF1 state 8 complete
          begin
          end

        6'h1F:
          begin
          end

        6'h20:
          begin
          end

        6'h21:
          begin
          end

        6'h22:
          begin
          end

        6'h23:
          begin
          end

        6'h24:
          begin
          end

        6'h25:
          begin
          end

        6'h26:
          begin
          end

        6'h27:
          begin
          end

        6'h28:
          begin
          end

        6'h29:
          begin
          end

        6'h2A:
          begin
          end

        6'h2B:
          begin
          end

        6'h2C:
          begin
          end

        6'h2D:
          begin
          end

        6'h2E:
          begin
          A0 <= PROD0;                      // lowpass SVF output
          B0 <= ADSRnca;                    // 
          end

        6'h2F:
          begin
          case (DACwin)
          4'h0: SynthOut <= PROD0[HIBIT-5:HIBIT-16];
          4'h1: SynthOut <= PROD0[HIBIT-6:HIBIT-17];
          4'h2: SynthOut <= PROD0[HIBIT-7:HIBIT-18];
          4'h3: SynthOut <= PROD0[HIBIT-8:HIBIT-19];
          4'h4: SynthOut <= PROD0[HIBIT-9:HIBIT-20];
          4'h5: SynthOut <= PROD0[HIBIT-10:HIBIT-21];
          4'h6: SynthOut <= PROD0[HIBIT-11:HIBIT-22];
          4'h7: SynthOut <= PROD0[HIBIT-12:HIBIT-23];
          4'h8: SynthOut <= PROD0[HIBIT-13:HIBIT-24];
          4'h9: SynthOut <= PROD0[HIBIT-14:HIBIT-25];
          4'hA: SynthOut <= PROD0[HIBIT-15:HIBIT-26];
          4'hB: SynthOut <= PROD0[HIBIT-16:HIBIT-27];
          4'hC: SynthOut <= PROD0[HIBIT-17:HIBIT-28];
          4'hD: SynthOut <= PROD0[HIBIT-18:HIBIT-29];
          4'hE: SynthOut <= PROD0[HIBIT-19:HIBIT-30];
          4'hF: SynthOut <= PROD0[HIBIT-20:HIBIT-31];
          endcase

          run <= 0;
          end

        endcase
        end
      end
    end


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////  
// Decode structures for hardware receiving data from the MCU

  always @ ( posedge clk )
  begin
    if ( write_strobe == 1'b1 )
    begin
// This case block contains selection logic for system level ports and for CC ports.
// Note that these ports all have the bit 7 or port_id set.
    case ( port_id )
//    8'hF0:       CHANpres              <= out_port[6:0];  // channel pressure value, global
//    8'hF1:       PW[13:7]              <= out_port[6:0];  // pitch wheel MSB, global
//    8'hF2:       PW[6:0]               <= out_port[6:0];  // pitch wheel LSB, global
//    8'hF3:       MOD_WHL               <= out_port[6:0];  // modulation wheel, global
    8'hF4:       NOTEOCT               <= out_port;         // F4 - F7 (note is low 4 bits)
    8'hF8:       MCU_GATE              <= out_port[0];    // MCU_GATE signal
//    8'hF9:       VEL                   <= out_port[6:0];  // port to set synth hardware velocity register
    8'hFC:       SUSTAIN               <= out_port[6];    // sustain command register
    8'hE0:       SYSEX_ADDR_MSB        <= out_port;
    8'hD0:       NOTENUM               <= out_port;       // this is used only as a modulation source
    8'h90:       LCD                   <= out_port;
    endcase
//////////////////////////////////////////////////////////////////////////////////////////
// These if blocks contain case blocks for sysex populated parameters     
// below this, all port_id values have bit 7 set to zero
// Address space is 14 bits.

    if ( SYSEX_ADDR_MSB == 8'h00 )
      begin
      case ( port_id )
      // SYSEX updatable parameter ports:    
      8'h10:       TRANSPOSE                       <= out_port[6:0];   // global transpose by semitones, neg values don't work - fix this
     
      8'h20:       NCAadsrA[13:7]                  <= out_port[6:0];
      8'h21:       NCAadsrA[6:0]                   <= out_port[6:0];
      8'h22:       NCAadsrD[13:7]                  <= out_port[6:0];
      8'h23:       NCAadsrD[6:0]                   <= out_port[6:0];
      8'h24:       NCAadsrS[13:7]                  <= out_port[6:0];
      8'h25:       NCAadsrS[6:0]                   <= out_port[6:0];
      8'h26:       NCAadsrR[13:7]                  <= out_port[6:0];
      8'h27:       NCAadsrR[6:0]                   <= out_port[6:0];
		 
      8'h30:       NCFadsrA[13:7]                  <= out_port[6:0];
      8'h31:       NCFadsrA[6:0]                   <= out_port[6:0];
      8'h32:       NCFadsrD[13:7]                  <= out_port[6:0];
      8'h33:       NCFadsrD[6:0]                   <= out_port[6:0];
      8'h34:       NCFadsrS[13:7]                  <= out_port[6:0];
      8'h35:       NCFadsrS[6:0]                   <= out_port[6:0];
      8'h36:       NCFadsrR[13:7]                  <= out_port[6:0];
      8'h37:       NCFadsrR[6:0]                   <= out_port[6:0];

      8'h40:       expo_R_NCA_ADSR                 <= out_port[0];
      8'h41:       expo_R_NCF_ADSR                 <= out_port[0];
      
      8'h7E:       DACwin                          <= out_port[3:0];   // DAC window value
       
      8'h7F:       MIDI_CHANNEL                    <= out_port[3:0];
      endcase
      end

    if ( SYSEX_ADDR_MSB == 8'h01 )
      begin
      case ( port_id )
      8'h00:       q_Scale[34:28]                  <= out_port[6:0];  // signed 35 bit
      8'h01:       q_Scale[27:21]                  <= out_port[6:0];
      8'h02:       q_Scale[20:14]                  <= out_port[6:0];
      8'h03:       q_Scale[13:7]                   <= out_port[6:0];
      8'h04:       q_Scale[6:0]                    <= out_port[6:0];

      8'h05:       q_Min[34:28]                    <= out_port[6:0];  // signed 35 bit
      8'h06:       q_Min[27:21]                    <= out_port[6:0];
      8'h07:       q_Min[20:14]                    <= out_port[6:0];
      8'h08:       q_Min[13:7]                     <= out_port[6:0];
      8'h09:       q_Min[6:0]                      <= out_port[6:0];

      8'h0A:       PmentoRate[34:28]               <= out_port[6:0];  // signed 35 bit
      8'h0B:       PmentoRate[27:21]               <= out_port[6:0];
      8'h0C:       PmentoRate[20:14]               <= out_port[6:0];
      8'h0D:       PmentoRate[13:7]                <= out_port[6:0];
      8'h0E:       PmentoRate[6:0]                 <= out_port[6:0];

      8'h0F:       NZlev[34:28]                    <= out_port[6:0];  // signed 35 bit
      8'h10:       NZlev[27:21]                    <= out_port[6:0];
      8'h11:       NZlev[20:14]                    <= out_port[6:0];
      8'h12:       NZlev[13:7]                     <= out_port[6:0];
      8'h13:       NZlev[6:0]                      <= out_port[6:0];

      8'h14:       Lowpass_q[34:28]                <= out_port[6:0];  // signed 35 bit
      8'h15:       Lowpass_q[27:21]                <= out_port[6:0];
      8'h16:       Lowpass_q[20:14]                <= out_port[6:0];
      8'h17:       Lowpass_q[13:7]                 <= out_port[6:0];
      8'h18:       Lowpass_q[6:0]                  <= out_port[6:0];

      8'h20:       Tuning[20:14]                   <= out_port[6:0];  // unsigned 21 bits comes from patch editor
      8'h21:       Tuning[13:7]                    <= out_port[6:0];
      8'h22:       Tuning[6:0]                     <= out_port[6:0];
 
      8'h25:       AmpComp[34:28]                  <= out_port[6:0];  // signed 35 bit
      8'h26:       AmpComp[27:21]                  <= out_port[6:0];
      8'h27:       AmpComp[20:14]                  <= out_port[6:0];
      8'h28:       AmpComp[13:7]                   <= out_port[6:0];
      8'h29:       AmpComp[6:0]                    <= out_port[6:0];
      endcase
      end       

//    if ( SYSEX_ADDR_MSB == 8'h02 )
//      begin
//      case ( port_id )
//      endcase
//      end       

//    if ( SYSEX_ADDR_MSB == 8'h03 )
//      begin
//      case ( port_id )
//      endcase
//      end       

    end
  end

// make sure that in_port_reg always contains selected data at rising edge of clk,
// PicoBlaze will read it when it needs it.
  always @ ( posedge clk ) 
    begin
    casex ( port_id[3:0] )                               // decode and transfer data to in_port_reg
    4'h0:    in_port_reg <= {rx1ready_status,rx0ready_status,6'b000000}; // UART1 & UART0 rxready bits
    4'h1:    in_port_reg <= rx0data;                     // MIDI UART rxdata
    4'h2:    in_port_reg <= {4'b0000,MIDI_CHANNEL};
    4'h8:    in_port_reg <= {1'b0,TRANSPOSE};
    4'h9:    in_port_reg <= rx1data;                     // TTY UART rxdata
    4'hF:    in_port_reg <= version;                     // The GateMan version number stored in hardware
    default: in_port_reg <= 8'bxxxxxxxx;
	  endcase
    end

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
// LEDs

assign LED = 8'h00;

 
//  wire [7:0] X;
//  assign X = {DIAG[34],|DIAG[33:30],|DIAG[29:25],|DIAG[24:20],|DIAG[19:15],|DIAG[14:10],|DIAG[9:5],|DIAG[4:0]};
//  assign LED = (BTN_W) ? X : {~Raw_MIDI_In,~TTY_In,MCU_GATE,SVF_oflo,MIDI_CHANNEL};

//assign LED = NOTENUMrev;


//assign X = {DIAG[34],|DIAG[33:30],|DIAG[29:25],|DIAG[24:20],|DIAG[19:15],|DIAG[14:10],|DIAG[9:5],|DIAG[4:0]};
//  assign LED = (BTN_W) ? X : {~Raw_MIDI_In,~TTY_In,MCU_GATE,SVF_oflo,MIDI_CHANNEL} ;


//assign LED = NOTEOCT;
//assign LED = (BTN_W) ? NOTENUM : NOTEOCT;

/////////////////////////////////////////////////////////////////////////////////////////////

endmodule